#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (C) Louisiana State University (2016, 2017)
#               Cardiff University (2017-2020)
#
# This file is part of PyOmicron.
#
# PyOmicron is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyOmicron is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#

"""Check the status of Omicron trigger generation
"""
import time
prog_start = time.time()

import argparse
import configparser
import json
import logging
import operator
import sys
import warnings
from collections import OrderedDict
from functools import reduce
from getpass import getuser
from os import linesep
from pathlib import Path
from textwrap import indent
from time import sleep

import htcondor

import numpy

from matplotlib import (rcParams, use)
from matplotlib.gridspec import GridSpec

import h5py

from MarkupPy import markup

from gwpy.io.cache import sieve as sieve_cache
from gwpy.time import to_gps
from gwpy.segments import (Segment, SegmentList)
from gwpy.plot import Plot
from gwpy.plot.segments import SegmentRectangle

from .. import (condor, const, io, log, segments, __version__)
from ..utils import get_omicron_version

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

NOW = int(to_gps('now'))
logger = None


def create_parser():
    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        '-V',
        '--version',
        action='version',
        version=__version__,
    )
    parser.add_argument(
        'group',
        help='name of channel group to check',
    )
    parser.add_argument(
        '-f',
        '--config-file', default=const.OMICRON_CHANNELS_FILE,
        help='path to a channel list configuration file',
    )
    parser.add_argument(
        '-i',
        '--ifo',
        default=const.IFO,
        help='IFO prefix to process',
    )
    parser.add_argument(
        '-s',
        '--gps-start-time',
        type=to_gps,
        default=NOW - 7 * 86400,
        help='GPS start time of check or date/time',
    )
    parser.add_argument(
        '-e',
        '--gps-end-time',
        type=to_gps,
        default=NOW,
        help='GPS end time of check or date/time',
    )
    parser.add_argument(
        '-c',
        '--channel',
        action='append',
        help='name of channel to process, can be given multiple '
             'times, default: all channels in group',
    )
    parser.add_argument(
        '-b',
        '--state-flag',
        help='name of data-quality flag to use in defining state, '
             'default is taken from --config-file',
    )
    parser.add_argument(
        '-p',
        '--state-pad',
        metavar='X,Y',
        default='0,0',
        help='inward padding for start,end of each '
             '--state-flag segment',
    )
    parser.add_argument(
        '-u',
        '--user',
        default=getuser(),
        help='name of user running condor jobs',
    )
    parser.add_argument(
        '-a',
        '--archive-directory',
        default=const.OMICRON_ARCHIVE,
        help='path of archive',
    )
    parser.add_argument(
        '-d',
        '--production-directory',
        default=const.OMICRON_PROD / "{group}",
        type=Path,
        help='path of production directory',
    )
    parser.add_argument(
        '-A',
        '--skip-condor',
        action='store_true',
        default=False,
        help="don't check condor status",
    )
    parser.add_argument(
        '-B',
        '--skip-file-checks',
        action='store_true',
        default=False,
        help="don't check file status",
    )
    parser.add_argument(
        '-C',
        '--skip-job-duration',
        action='store_true',
        default=False,
        help="don't query and plot job durations",
    )

    pout = parser.add_argument_group('Output options')
    pout.add_argument(
        '--json',
        const=True,
        nargs='?',
        default=False,
        help='print output in dashboard.ligo.org nagios JSON format',
    )
    pout.add_argument(
        '-o',
        '--output-directory',
        default=Path.cwd(),
        type=Path,
        help='path to write output',
    )
    pout.add_argument(
        '-l',
        '--latency-archive-tag',
        default='{group}',
        help='file tag for latency archive',
    )
    pout.add_argument(
        '-m',
        '--html',
        default=False,
        action='store_true',
        help='write HTML summary to index.html in output dir',
    )

    pnag = parser.add_argument_group('Monitoring options')
    pnag.add_argument(
        '-U',
        '--unknown',
        type=int,
        default=7200,
        help='time (seconds) after which nagios output should be '
             'considered stable and \'unknown\'',
    )
    pnag.add_argument(
        '-W',
        '--warning',
        type=int,
        default=3600,
        help='how much latency to consider as a warning',
    )
    pnag.add_argument(
        '-X',
        '--error',
        type=int,
        default=3600 * 2,
        help='how much latency to consider as an error',
    )
    return parser


def main(args=None):
    global logger
    use('agg')
    rcParams.update({
        'figure.subplot.bottom': 0.15,
        'figure.subplot.left': 0.1,
        'figure.subplot.right': 0.83,
        'figure.subplot.top': 0.93,
        'figure.subplot.hspace': 0.25,
        'axes.labelsize': 20,
        'grid.color': 'gray',
    })
    grid = GridSpec(2, 1)

    logger = log.Logger('omicron-status')
    logger.setLevel(logging.DEBUG)

    try:
        omicronversion = str(get_omicron_version())
    except KeyError:
        omicronversion = 'Unknown'
        logger.warning("Omicron version unknown")
    else:
        logger.info("Found omicron version: %s" % omicronversion)

    parser = create_parser()
    args = parser.parse_args(args=args)

    if args.ifo is None:
        parser.error("Cannot determine IFO prefix from system, "
                     "please pass --ifo on the command line")

    group = args.group

    logger.info("Checking status for %r group" % group)

    archive = args.archive_directory
    proddir = args.production_directory.with_name(
        args.production_directory.name.format(group=args.group),
    )
    outdir = args.output_directory
    outdir.mkdir(exist_ok=True, parents=True)
    tag = args.latency_archive_tag.format(group=args.group)

    filetypes = ['h5', 'xml.gz', 'root']

    logger.debug("Set output directory to %s" % outdir)
    logger.debug(
        "Will process the following filetypes: {}".format(
            ", ".join(filetypes),
        ),
    )

    # -- parse configuration file and get parameters --------------------------
    ifo = args.ifo
    if ifo is None:
        raise ValueError('IFO is unknown.')
    if args.config_file is None:
        config_file = const.OMICRON_PROD / f"{ifo}-channels.ini"
    else:
        config_file = args.config_file
    config_file = Path(config_file)
    if not config_file.exists():
        raise IOError(f'Configuration file does not exist: {str(config_file.absolute())}')
    cp = configparser.ConfigParser()
    ok = cp.read(config_file)
    if str(config_file) not in ok:
        raise IOError(f"Failed to read configuration file {str(config_file.absolute())}")
    logger.info("Configuration read")

    # validate
    if not cp.has_section(group):
        raise configparser.NoSectionError(group)

    # get parameters
    obs = args.ifo[0]
    frametype = cp.get(group, 'frametype')
    padding = cp.getint(group, 'overlap-duration') / 2.
    mingap = cp.getint(group, 'chunk-duration')

    channels = args.channel
    if not channels:
        channels = [c.split()[0] for
                    c in cp.get(group, 'channels').strip('\n').split('\n')]
    channels.sort()
    logger.debug("Found %d channels" % len(channels))

    start = args.gps_start_time
    end = args.gps_end_time
    if end == NOW:
        end -= padding

    if args.state_flag:
        stateflag = args.state_flag
        statepad = tuple(map(float, args.state_pad.split(',')))
    else:
        try:
            stateflag = cp.get(group, 'state-flag')
        except configparser.NoOptionError:
            stateflag = None
        else:
            try:
                statepad = tuple(map(
                    float,
                    cp.get(group, 'state-padding').split(','),
                ))
            except configparser.NoOptionError:
                statepad = (0, 0)
    if stateflag:
        logger.debug("Parsed state flag: %r" % stateflag)
        logger.debug("Parsed state padding: %s" % repr(statepad))
    logger.info("Processing %d-%d" % (start, end))

    # -- define nagios JSON printer -------------------------------------------

    def print_nagios_json(code, message, outfile, tag='status', **extras):
        out = {
            'created_gps': NOW,
            'status_intervals': [
                {'start_sec': 0,
                 'end_sec': args.unknown,
                 'num_status': code,
                 'txt_status': message},
                {'start_sec': args.unknown,
                 'num_status': 3,
                 'txt_status': 'Omicron %s check is not running' % tag},
            ],
            'author': {
                'name': 'Duncan Macleod',
                'email': 'duncan.macleod@ligo.org',
            },
            'omicron': {
                'version': omicronversion,
                'group': group,
                'channels': ' '.join(channels),
                'frametype': frametype,
                'state-flag': stateflag,
            },
            'pyomicron': {
                'version': __version__,
            },
        }
        out.update(extras)
        with open(outfile, 'w') as f:
            f.write(json.dumps(out, indent=3))
        logger.debug("nagios info written to %s" % outfile)

    # -- get condor status ------------------------------------------------

    if not args.skip_condor:
        # connect to scheduler
        try:
            schedd = htcondor.Schedd()
        except RuntimeError as e:
            logger.warning("Caught %s: %s" % (type(e).__name__, e))
            logger.info("Failed to connect to HTCondor scheduler, cannot "
                        "determine condor status for %s" % group)
            schedd = None

    if not args.skip_condor and schedd:
        logger.info("-- Checking condor status --")

        # get DAG status
        jsonfp = outdir / "nagios-condor-{}.json".format(group)
        okstates = ['Running', 'Idle', 'Completed']
        try:
            # check manager status
            qstr = 'OmicronManager == "{}" && Owner == "{}"'.format(
                group,
                args.user,
            )
            try:
                jobs = schedd.query(qstr, ['JobStatus'])
            except IOError as e:
                warnings.warn("Caught IOError: %s [retrying...]" % str(e))
                sleep(2)
                jobs = schedd.query(qstr, ['JobStatus'])
            logger.debug(
                "Found {} jobs for query {!r}".format(len(jobs), qstr),
            )
            if len(jobs) > 1:
                raise RuntimeError("Multiple OmicronManager jobs found for %r"
                                   % group)
            elif len(jobs) == 0:
                raise RuntimeError(
                    "No OmicronManager job found for %r" % group,
                )
            status = condor.JOB_STATUS[jobs[0]['JobStatus']]
            if status not in okstates:
                raise RuntimeError("OmicronManager status for %r: %r"
                                   % (group, status))
            logger.debug("Manager status is %r" % status)
            # check node status
            jobs = schedd.query(
                'OmicronProcess == "{}" && Owner == "{}"'.format(
                    group,
                    args.user,
                ),
                ['JobStatus', 'ClusterId'],
            )
            logger.debug(
                "Found {} jobs for query {!r}".format(len(jobs), qstr),
            )
            for job in jobs:
                status = condor.JOB_STATUS[job['JobStatus']]
                if status not in okstates:
                    raise RuntimeError("Omicron node %s (%r) is %r"
                                       % (job['ClusterId'], group, status))
        except RuntimeError as e:
            print_nagios_json(2, str(e), jsonfp, tag='condor')
            logger.warning("Failed to determine condor status: %r" % str(e))
        except IOError as e:
            logger.warning("Caught %s: %s" % (type(e).__name__, e))
            logger.info("Failed to connect to HTCondor scheduler, cannot "
                        "determine condor status for %s" % group)
        else:
            print_nagios_json(
                0,
                "Condor processing for %r is OK" % group,
                jsonfp,
                tag='condor',
            )
            logger.info("Condor processing is OK")

    if not args.skip_job_duration:
        # get job duration history
        plot = Plot(figsize=[12, 3])
        plot.subplots_adjust(bottom=.22, top=.87)
        ax = plot.gca(xscale="auto-gps")
        times, jobdur = condor.get_job_duration_history_shell(
            'OmicronProcess', group, maxjobs=5000)
        logger.debug("Recovered duration history for %d omicron.exe jobs"
                     % len(times))
        line = ax.plot([0], [1], label='Omicron.exe')[0]
        ax.plot(times, jobdur, linestyle=' ', marker='.',
                color=line.get_color())
        times, jobdur = condor.get_job_duration_history_shell(
            'OmicronPostProcess', group, maxjobs=5000)
        logger.debug("Recovered duration history for %d post-processing jobs"
                     % len(times))
        line = ax.plot([0], [1], label='Post-processing')[0]
        ax.plot(times, jobdur, linestyle=' ', marker='.',
                color=line.get_color())
        ax.legend(loc='upper left', borderaxespad=0, bbox_to_anchor=(1.01, 1),
                  handlelength=1)
        ax.set_xlim(args.gps_start_time, args.gps_end_time)
        ax.set_epoch(ax.get_xlim()[1])
        ax.set_yscale('log')
        ax.set_title('Omicron job durations for %r' % group)
        ax.set_ylabel('Job duration [seconds]')
        ax.xaxis.labelpad = 5
        png = str(outdir / "nagios-condor-{}.png".format(group))
        plot.save(png)
        plot.close()
        logger.debug("Saved condor plot to %s" % png)

    if args.skip_file_checks:
        sys.exit(0)

    # -- get file latency and archive completeness ----------------------------

    logger.info("-- Checking file archive --")

    # get frame segments
    segs = segments.get_frame_segments(obs, frametype, start, end)

    # get state segments
    if stateflag is not None:
        segs &= segments.query_state_segments(
            stateflag,
            start,
            end,
            pad=statepad,
        )

    try:
        end = segs[-1][1]
    except IndexError:
        pass

    # apply inwards padding to generate resolvable segments
    for i in range(len(segs) - 1, -1, -1):
        # if segment is shorter than padding, ignore it completely
        if abs(segs[i]) <= padding * 2:
            del segs[i]
        # otherwise apply padding to generate trigger segment
        else:
            segs[i] = segs[i].contract(padding)
    logger.debug("Found %d seconds of analysable time" % abs(segs))

    # load archive latency
    latencyfile = outdir / "nagios-latency-{}.h5".format(tag)
    times = dict((c, dict((ft, None) for ft in filetypes)) for c in channels)
    ldata = dict((c, dict((ft, None) for ft in filetypes)) for c in channels)
    try:
        with h5py.File(latencyfile, 'r') as h5file:
            for c in channels:
                for ft in filetypes:
                    try:
                        times[c][ft] = h5file[c]['time'][ft][:]
                        ldata[c][ft] = h5file[c]['latency'][ft][:]
                    except KeyError:
                        times[c][ft] = numpy.ndarray((0,))
                        ldata[c][ft] = numpy.ndarray((0,))
    except OSError as exc:  # file not found, or is corrupt
        warnings.warn("failed to load latency data from {}: {}".format(
            latencyfile,
            str(exc),
        ))
        for c in channels:
            for ft in filetypes:
                if not times[c].get(ft):
                    times[c][ft] = numpy.ndarray((0,))
                    ldata[c][ft] = numpy.ndarray((0,))
    else:
        logger.debug("Parsed latency data from %s" % latencyfile)

    # load acknowledged gaps
    acksegfile = str(outdir / "acknowledged-gaps-{}.txt".format(tag))
    try:
        acknowledged = SegmentList.read(acksegfile, gpstype=float,
                                        format="segwizard")
    except IOError:  # no file
        acknowledged = SegmentList()
    else:
        logger.debug(
            "Read %d segments from %s" % (len(acknowledged), acksegfile),
        )
        acknowledged.coalesce()

    # build legend for segments
    leg = OrderedDict()
    leg['Analysable'] = SegmentRectangle(
        [0, 1], 0, facecolor='lightgray', edgecolor='gray',
    )
    leg['Available'] = SegmentRectangle(
        [0, 1], 0, facecolor='lightgreen', edgecolor='green',
    )
    leg['Missing'] = SegmentRectangle(
        [0, 1], 0, facecolor='red', edgecolor='darkred',
    )
    leg['Unresolvable'] = SegmentRectangle(
        [0, 1], 0, facecolor='magenta', edgecolor='purple',
    )
    leg['Overlapping'] = SegmentRectangle(
        [0, 1], 0, facecolor='yellow', edgecolor='orange',
    )
    leg['Pending'] = SegmentRectangle(
        [0, 1], 0, facecolor='lightskyblue', edgecolor='blue',
    )
    leg['Acknowledged'] = SegmentRectangle(
        [0, 1], 0, facecolor='sandybrown', edgecolor='brown',
    )

    logger.debug("Checking archive for each channel...")

    # find files
    latency = {}
    gaps = {}
    overlap = {}
    pending = {}
    plots = {}
    for c in channels:
        # create data storage
        latency[c] = {}
        gaps[c] = {}
        overlap[c] = {}
        pending[c] = {}

        # create figure
        plot = Plot(figsize=[12, 5])
        lax = plot.add_subplot(grid[0, 0], xscale="auto-gps")
        sax = plot.add_subplot(grid[1, 0], sharex=lax, projection='segments')
        colors = ['lightblue', 'dodgerblue', 'black']

        for y, ft in enumerate(filetypes):
            # find files
            cache = io.find_omicron_files(c, start, end, archive, ext=ft)
            cpend = sieve_cache(io.find_pending_files(c, proddir, ext=ft),
                                segment=Segment(start, end))
            # get available segments
            avail = segments.cache_segments(cache)
            found = avail & segs
            pending[c][ft] = segments.cache_segments(cpend) & segs
            # remove gaps at the end that represent latency
            try:
                latency[c][ft] = abs(segs & type(segs)([
                    type(segs[0])(found[-1][1], segs[-1][1])])) / 3600.
            except IndexError:
                latency[c][ft] = 0
                processed = segs
            else:
                processed = segs & type(segs)(
                    [type(segs[0])(start, found[-1][1])])
            gaps[c][ft] = type(found)()
            lost = type(found)()
            for s in processed - found:
                if abs(s) < mingap and s in list(segs):
                    lost.append(s)
                else:
                    gaps[c][ft].append(s)
            # remove acknowledged gaps
            ack = gaps[c][ft] & acknowledged
            gaps[c][ft] -= acknowledged
            # print warnings
            if abs(gaps[c][ft]):
                warnings.warn(
                    f"Gaps found in {c} files for {ft}:\n{gaps[c][ft]}")
            overlap[c][ft] = segments.cache_overlaps(cache)
            if abs(overlap[c][ft]):
                warnings.warn(
                    f"Overlap found in {c} files for {ft}:\n{overlap[c][ft]}")

            # append archive
            times[c][ft] = numpy.concatenate((times[c][ft][-99999:], [NOW]))
            ldata[c][ft] = numpy.concatenate((ldata[c][ft][-99999:],
                                              [latency[c][ft]]))

            # plot
            line = lax.plot(
                times[c][ft],
                ldata[c][ft],
                label=ft,
                color=colors[y],
            )[0]
            lax.plot(times[c][ft], ldata[c][ft], marker='.', linestyle=' ',
                     color=line.get_color())
            sax.plot_segmentlist(segs, y=y, label=ft, alpha=.5,
                                 facecolor=leg['Analysable'].get_facecolor(),
                                 edgecolor=leg['Analysable'].get_edgecolor())
            sax.plot_segmentlist(pending[c][ft], y=y,
                                 facecolor=leg['Pending'].get_facecolor(),
                                 edgecolor=leg['Pending'].get_edgecolor())
            sax.plot_segmentlist(avail, y=y, label=ft, alpha=.2, height=.1,
                                 facecolor=leg['Available'].get_facecolor(),
                                 edgecolor=leg['Available'].get_edgecolor())
            sax.plot_segmentlist(found, y=y, label=ft, alpha=.5,
                                 facecolor=leg['Available'].get_facecolor(),
                                 edgecolor=leg['Available'].get_edgecolor())
            sax.plot_segmentlist(lost, y=y,
                                 facecolor=leg['Unresolvable'].get_facecolor(),
                                 edgecolor=leg['Unresolvable'].get_edgecolor())
            sax.plot_segmentlist(gaps[c][ft], y=y,
                                 facecolor=leg['Missing'].get_facecolor(),
                                 edgecolor=leg['Missing'].get_edgecolor())
            sax.plot_segmentlist(overlap[c][ft], y=y,
                                 facecolor=leg['Overlapping'].get_facecolor(),
                                 edgecolor=leg['Overlapping'].get_edgecolor())
            sax.plot_segmentlist(ack, y=y,
                                 facecolor=leg['Acknowledged'].get_facecolor(),
                                 edgecolor=leg['Acknowledged'].get_edgecolor())

        # finalise plot
        lax.axhline(args.warning / 3600., color=(1.0, 0.7, 0.0),
                    linestyle='--', linewidth=2, label='Warning', zorder=-1)
        lax.axhline(args.error / 3600., color='red', linestyle='--',
                    linewidth=2, label='Critical', zorder=-1)
        lax.set_title('Omicron status: {}'.format(c))
        lax.set_ylim(0, args.error / 1800.)
        lax.set_ylabel('Latency [hours]')
        lax.legend(loc='upper left', bbox_to_anchor=(1.01, 1), borderaxespad=0,
                   handlelength=2, fontsize=12.4)
        lax.set_xlabel(' ')
        for ax in plot.axes:
            ax.set_xlim(args.gps_start_time, args.gps_end_time)
            ax.set_epoch(ax.get_xlim()[1])
        sax.xaxis.labelpad = 5
        sax.set_ylim(-.5, len(filetypes) - .5)
        sax.legend(leg.values(), leg.keys(), handlelength=1, fontsize=12.4,
                   loc='lower left', bbox_to_anchor=(1.01, 0), borderaxespad=0)
        plots[c] = png = outdir / "nagios-latency-{}.png".format(
            c.replace(':', '-'),
        )
        plot.save(png)
        plot.close()
        logger.debug("    %s" % c)

    # update latency and write archive
    h5file = h5py.File(latencyfile, 'w')
    for c in channels:
        g = h5file.create_group(c)
        for name, d in zip(['time', 'latency'], [times[c], ldata[c]]):
            g2 = g.create_group(name)
            for ft in filetypes:
                g2.create_dataset(ft, data=d[ft], compression='gzip')
    h5file.close()
    logger.debug("Stored latency data as HDF in %s" % latencyfile)

    # write nagios output for files
    status = []
    for segset, tag in zip([gaps, overlap], ['gaps', 'overlap']):
        chans = [(c, segset[c]) for c in segset
                 if abs(reduce(operator.or_, segset[c].values()))]
        jsonfp = outdir / "nagios-{}-{}.json".format(tag, group)
        status.append((tag, jsonfp))
        if chans:
            # format the segments nicely for Icinga 2
            lines = []
            for chan, ftsegs in chans:
                lines.append(f"{chan}:")
                for ft, segs in ftsegs.items():
                    lines.extend((
                        f"  {ft}:",
                        indent(str(segs), "    "),
                    ))
            gapstr = linesep.join(lines)
            code = 1
            message = ("%s found in Omicron files for group %r\n%s"
                       % (tag.title(), group, gapstr))
        else:
            code = 0
            message = (
                "No %s found in Omicron files for group %r" % (tag, group)
            )
        print_nagios_json(code, message, jsonfp, tag=tag, **{tag: dict(chans)})

    # write group JSON
    jsonfp = outdir / "nagios-latency-{}.json".format(group)
    status.append(('latency', jsonfp))
    code = 0
    message = 'No channels have high latency for group %r' % group
    ldict = dict((c, max(latency[c].values())) for c in latency)
    for x, dt in zip([2, 1], [args.error, args.warning]):
        dh = dt / 3600.
        chans = [c for c in ldict if ldict[c] >= dh]
        if chans:
            code = x
            message = ("%d channels found with high latency (above %s seconds)"
                       % (len(chans), dt))
            break
    print_nagios_json(code, message, jsonfp, tag='latency', latency=ldict)

    # auto-detect 'standard' JSON files
    for tag, name in zip(
        ['condor', 'omicron-online'],
        ['condor', 'processing'],
    ):
        f = outdir / "nagios-{}-{}.json".format(tag, group)
        if f.is_file():
            status.insert(0, (name, f))

    # write HTML summary
    if args.html:
        page = markup.page()
        page.init(
            title="%s Omicron Online status" % group,
            css=[
                ('//maxcdn.bootstrapcdn.com/bootstrap/3.3.4/css/'
                 'bootstrap.min.css'),
                ('//cdnjs.cloudflare.com/ajax/libs/fancybox/2.1.5/'
                 'jquery.fancybox.min.css'),
            ],
            script=[
                '//code.jquery.com/jquery-1.11.2.min.js',
                ('//maxcdn.bootstrapcdn.com/bootstrap/3.3.4/js/'
                 'bootstrap.min.js'),
                ('//cdnjs.cloudflare.com/ajax/libs/fancybox/2.1.5/'
                 'jquery.fancybox.min.js'),
            ],
        )
        page.div(class_='container')
        # write header
        page.div(class_='page-header')
        page.h1('Omicron Online status: %s' % group)
        page.div.close()  # page-header
        # write summary
        page.div(id_='json')
        page.h2("Processing status")
        for tag, f in status:
            jf = f.name
            page.a("%s status" % tag.title(), href=jf, role='button',
                   target="_blank", id_="nagios-%s" % tag,
                   class_='btn btn-default json-status')
        page.p(style="padding-top: 5px;")
        page.small(
            "Hover over button for explanation, click to open JSON file",
        )
        page.p.close()
        page.div.close()  # id=json
        # show plots
        page.div(id_='plots')
        page.h2("Channel details")
        page.div(class_='row')
        for channel in sorted(channels):
            png = plots[channel].name
            page.div(class_="col-sm-6 col-md-4")
            page.div(class_="panel panel-default")
            page.div(class_='panel-heading')
            page.h3(channel, class_='panel-title', style="font-size: 14px;")
            page.div.close()  # panel-heading
            page.div(class_='panel-body')
            page.a(href=png, target="_blank", class_="fancybox",
                   rel="channel-status-img")
            page.img(src=png, class_='img-responsive')
            page.a.close()
            page.div.close()  # panel-body
            page.div.close()  # panel
            page.div.close()  # col
        page.div.close()  # row
        page.div.close()  # id=plots

        # dump parameters
        page.div(id_="parameters")
        page.h2("Parameters")
        for key, val in cp.items(group):
            page.p()
            page.strong("%s:" % key)
            page.add(val)
            page.p.close()
        page.div.close()  # id=parameters

        # finish and close
        page.div.close()  # container
        page.script("""
        function setStatus(data, id) {
            var txt = data.status_intervals[0].txt_status.split("\\n")[0];
            $("#"+id).attr("title", txt);
            var stat = data.status_intervals[0].num_status;
            if (stat == 0) {
                $("#"+id).addClass("btn-success"); }
            else if (stat == 1) {
                $("#"+id).addClass("btn-warning"); }
            else if (stat == 2){
                $("#"+id).addClass("btn-danger"); }
        }

        $(document).ready(function() {
            $(".json-status").each(function() {
                var jsonf = $(this).attr("href");
                var id = $(this).attr("id");
                $.getJSON(jsonf, function(data) { setStatus(data, id); });
            });

            $(".fancybox").fancybox({nextEffect: 'none', prevEffect: 'none'});
        });""", type="text/javascript")
        with (outdir / "index.html").open("w") as f:
            f.write(str(page))
        logger.debug("HTML summary written to %s" % f.name)


if __name__ == "__main__":
    main()
    if logger:
        logger.info(f'Run time: {(time.time()-prog_start):.1f} seconds')



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2016)
#
# This file is part of PyOmicron.
#
# PyOmicron is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyOmicron is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyOmicron.  If not, see <http://www.gnu.org/licenses/>.

"""Read/write/modify Omicron-format parameters files
"""

import configparser
import os.path
import re
from collections import OrderedDict
from datetime import datetime
from getpass import getuser
from math import (ceil, exp, floor, log)

from ligo.segments import (segmentlist as SegmentList, segment as Segment)

from . import (const, utils)
from .segments import integer_segments

CHANNEL_DELIM_REGEX = re.compile(r'[:_-]')
UNUSED_PARAMS = ['state-flag', 'frametype', 'state-channel', 'state-frametype']
OMICRON_PARAM_MAP = {
    'sample-frequency': ('DATA', 'SAMPLEFREQUENCY')
}


class OmicronParameters(configparser.ConfigParser):
    """Custom `configparser.ConfigParser` for Omicron parameters files
    """
    OMICRON_DEFAULTS = OrderedDict()
    OMICRON_DEFAULTS[None] = {
        'PARAMETER': {
            'CLUSTERING': 'TIME',
            'FFTPLAN': 'FFTW_ESTIMATE',
            'TRIGGERRATEMAX': 100000,
        },
        'OUTPUT': {
            'DIRECTORY': os.path.curdir,
            'PRODUCTS': 'triggers',
            'VERBOSITY': 1,
            'FORMAT': 'root xml hdf5',
            'NTRIGGERMAX': 1e7,
        },
        'DATA': {
        },
    }

    def __init__(self, version=None, defaults=dict(), **kwargs):
        configparser.ConfigParser.__init__(self, defaults=defaults, **kwargs)
        if version is None:
            try:
                version = utils.get_omicron_version()
            except KeyError:
                version = utils.OmicronVersion(const.OMICRON_VERSION)
        self.version = version
        self._set_defaults()

    def _set_defaults(self):
        """Set basic defaults for each config section
        """
        for version in self.OMICRON_DEFAULTS:
            if version is not None and self.version < version:
                continue
            for section, params in self.OMICRON_DEFAULTS[version].items():
                try:
                    self.add_section(section)
                except configparser.DuplicateSectionError:
                    pass
                for key, val in params.items():
                    if isinstance(val, tuple):
                        self.set(section, key, ' '.join(map(str, val)))
                    else:
                        self.set(section, key, str(val))

    # -- better option accessors ----------------

    def getlist(self, section, option):
        raw = self.get(section, option)
        return raw.split()

    def getfloats(self, section, option):
        return list(map(float, self.getlist(section, option)))

    def optionxform(self, optionstr):
        return optionstr.upper()

    # -- input/output ---------------------------

    def _read(self, fp, fpname):
        """Read a file either either using INI or Omicron formatting
        """
        if fpname.endswith('.ini'):
            return configparser.ConfigParser._read(self, fp, fpname)
        channels = []
        for line in fp:
            if isinstance(line, bytes):
                line = line.decode()
            if not line.strip() or line[0] in '#;':  # blank
                continue
            sec, key, val = line.rstrip().split(None, 2)
            if sec.lower() == 'data' and key.lower() == 'channels':
                channels.append(val)
            else:
                self.set(sec, key, val)
        if channels:
            self.set('DATA', 'CHANNELS', ' '.join(channels))

    def write(self, fp):
        # write basic
        if fp.name.endswith('.ini'):
            return configparser.ConfigParser.write(self, fp)

        # print header
        print('# Omicron %s parameter file' % self.version, file=fp)
        print('# Written using pyomicron by %s at %s'
              % (getuser(), datetime.now()), file=fp)

        # write omicron-format
        def _write_option(sec, key, val):
            print('{0: <10}'.format(sec.upper()),
                  '{0: <16}'.format(key.upper()),
                  val, file=fp, sep=' ')

        for sec in self.sections():
            print("", file=fp)
            for key, val in self.items(sec):
                if sec.lower() == 'data' and key.lower() == 'channels':
                    for chan in val.split():
                        _write_option(sec, key, chan)
                else:
                    _write_option(sec, key, val)
    write.__doc__ = configparser.ConfigParser.write.__doc__

    def write_distributed(self, directory, nchannels=10):
        """Write multiple parameters files to distribute processing

        This method separates the channels list into blocks of ``nchannels``
        and writes a parameters file for each one into ``directory``.

        Parameters
        ----------
        directory : `str`
            the output directory for the parameters files

        nchannels : `int`, optional, default: ``10``
            the number of channels to insert in each parameters file

        Returns
        -------
        paramfile : `str`
            the file path of the complete parameters.txt file, written for
            reference
        parametermap : `dict`
            a `dict` of ``(filename, channel_list)`` pairs that indicate
            which written file contains which channels
        """
        # get channels list
        channels = self.getlist('DATA', 'CHANNELS')
        with open(os.path.join(directory, 'parameters.txt'), 'w') as f:
            self.write(f)
        tmpcp = self.__class__(version=self.version)
        tmpcp.read([f.name])

        # write files
        files = OrderedDict()
        i = 0
        while i * nchannels < len(channels):
            chans = channels[i * nchannels:(i + 1) * nchannels]
            pfile = os.path.join(directory, 'parameters-%d.txt' % i)
            tmpcp.set('DATA', 'CHANNELS', ' '.join(chans))
            with open(pfile, 'w') as f2:
                tmpcp.write(f2)
            files[pfile] = chans
            i += 1
        return f.name, files

    @classmethod
    def from_channel_list_config(cls, config, section, version=None):
        """Read `OmicronParameters` from a LIGO channel list file configuration

        Parameters
        ----------
        config : `configparser.ConfigParser`, `str`
            either a configuration object, or a path to a config file

        section : `str`
            the name of the section to parse

        version : `str`, optional
           the version of Omicron for which this configuration is valid

        Returns
        -------
        parameters : `OmicronParameters`
            an Omicron-format version of the parameters from the configuration
        """
        new = cls(version=version)

        # open and parse config filename
        if isinstance(config, str):
            fn = config
            config = configparser.ConfigParser()
            config.read(fn)

        # reformat from separate low, high to a tuple of values
        for (range_, low_, high_) in [('frequency-range', 'flow', 'fhigh'),
                                      ('q-range', 'qlow', 'qhigh')]:
            if not config.has_option(section, range_) and config.has_option(section, low_):
                config.set(
                    section, range_,
                    '%s %s' % (config.getfloat(section, low_),
                               config.getfloat(section, high_)))
            for opt in (low_, high_):
                try:
                    config.remove_option(section, opt)
                except configparser.NoOptionError:
                    pass

        for key, val in config.items(section):
            if key in UNUSED_PARAMS:
                continue
            elif key == 'channels':
                new.set('DATA', 'CHANNELS', ' '.join(val.split()))
            else:
                omikey = ''.join(key.split('-')).upper()
                try:
                    group, attr = OMICRON_PARAM_MAP[key]
                except KeyError:
                    if val in [None, 'None', 'none', '']:  # unset default
                        new.remove_option('PARAMETER', omikey)
                    else:
                        new.set('PARAMETER', omikey, val)
                else:
                    new.set(group, omikey, val)

        # -- get parameters
        # rename 'CHUNKDURATION' -> 'PSDLENGTH'
        try:
            psdlength = new.get('PARAMETER', 'CHUNKDURATION')
        except configparser.NoOptionError:
            pass
        else:
            new.remove_option('PARAMETER', 'CHUNKDURATION')
            new.set('PARAMETER', 'PSDLENGTH', psdlength)
        # combine 'SEGMENTDURATION' and 'OVERLAPDURATION' into 'TIMING'
        try:
            timing = (new.get('PARAMETER', 'SEGMENTDURATION'),
                      new.get('PARAMETER', 'OVERLAPDURATION'))
        except configparser.NoOptionError:
            pass
        else:
            new.remove_option('PARAMETER', 'SEGMENTDURATION')
            new.remove_option('PARAMETER', 'OVERLAPDURATION')
            new.set('PARAMETER', 'TIMING', '%s %s' % timing)

        return new

    # -- utilities ------------------------------

    def validate(self):
        """Validate that Omicron will accept the segment parameters

        Parameters
        ----------
        chunk : `int`
            omicron CHUNKDURATION parameter
        segment : `int`
            omicron SEGMENTDURATION parameter
        overlap : `int`
            omicron OVERLAPDURATION parameter
        frange : `tuple` of `float`
            omicron FREQUENCYRANGE parameter

        Raises
        ------
        AssertionError
            if any of the parameters isn't acceptable for Omicron to process
        """
        # extract parameters (for convenience)
        try:
            sampling = self.getfloat('DATA', 'SAMPLEFREQUENCY')
        except configparser.NoOptionError:
            sampling = None
        try:
            frange = self.getfloats('PARAMETER', 'FREQUENCYRANGE')
        except configparser.NoOptionError:
            frange = None
        chunk = self.getfloat('PARAMETER', 'PSDLENGTH')
        segment, overlap = self.getfloats('PARAMETER', 'TIMING')

        # check parameters are valid
        assert segment <= chunk, (
            "Segment length is greater than chunk length")
        assert overlap <= segment, (
            "Overlap length is greater than segment length")
        assert overlap != 0, "Omicron cannot run with zero overlap"
        assert overlap % 2 == 0, "Padding (overlap/2) is non-integer"
        assert segment >= (2 * overlap), (
            "Overlap is too large, cannot be more than 50%")
        dchunk = chunk - overlap
        dseg = segment - overlap
        assert dchunk % dseg == 0, (
            "Chunk duration doesn't allow an integer number of segments, "
            "%ds too large" % (dchunk % dseg))
        if sampling is None or frange is None:
            return
        if frange[0] < 1:
            x = 10 * floor(sampling / frange[0])
            psdsize = 2 * int(2 ** ceil(log(x) / log(2.)))
        else:
            psdsize = 2 * sampling
        psdlen = psdsize / sampling
        chunkp = chunk * sampling
        overlapp = overlap * sampling
        flow = 5 * sampling / exp(log((chunk - overlap) / 4., 2))
        assert (chunkp - overlapp) >= 2 * psdsize, (
            "Chunk duration not large enough to resolve lower-frequency "
            "bound, Omicron needs at least %ds. Minimum lower-frequency bound "
            "for this chunk duration is %.2gHz" % (2 * psdlen + overlap, flow))

    @integer_segments
    def output_segments(self, start, end):
        """Prints the list of processing segments for a given data segment
        """
        # get parameters
        segment, overlap = self.getfloats('PARAMETER', 'TIMING')
        fileduration = segment - overlap
        padding = overlap / 2.

        # build list of file segments
        t = start + padding
        stop = end - padding
        segments = SegmentList()
        while t < stop:
            e = min(t + fileduration, stop)
            segments.append(Segment(t, e))
            t = e
        return segments

    @integer_segments
    def distribute_segment(self, start, end, nperjob=1):
        """Determine processing segments to parallelise an Omicron job

        This function is meant to return a `segmentlist` of job [start, stop)
        times to pass to condor. Each segment will have duration
        `chunk * nperjob` *OR* `chunk * nperjob + remainder` if the remainder
        until the `end` is less than 1 chunk.

        Parameters
        ----------
        start : `int`
            the start GPS time of the Omicron segment

        end : `int`
            the end GPS time of the Omicron segment

        nperjob : `int`
            the number of chunks to put into each job

        Returns
        -------
        jobsegs : :class:`~ligo.segments.segmentlist`
            a `segmentlist` of [start, stop) times over which to distribute
            a single segment under condor
        """
        # get parameters
        chunk = self.getfloat('PARAMETER', 'PSDLENGTH')
        overlap = self.getfloats('PARAMETER', 'TIMING')[1]

        # single small segment
        if end - start <= chunk * 2:
            return SegmentList([Segment(start, end)])

        # multiple larger segments
        out = SegmentList()
        t = start
        while t < end - overlap:
            seg = Segment(t, t)
            c = chunk
            while abs(seg) < chunk * nperjob and seg[1] < end:
                seg = Segment(seg[0], min(seg[1] + c, end))
                c = chunk - overlap
            if abs(seg) < chunk:
                out[-1] += seg
            else:
                out.append(seg)
            t = seg[1] - overlap
        return out

    def output_formats(self):
        return [fmt for fmt in ('root', 'txt', 'xml', 'hdf5') if
                fmt in self.get('OUTPUT', 'FORMAT')]

    def output_files(self, start, end, flatten=False):
        """Prints the list of output files for a given processing segment

        Parameters
        ----------
        start : `int`
            GPS start time of job

        end : `int`
            GPS end time of job

        flatten : `bool`, optional, default: `False`
            return a flat list of file names, rather than the default
            `dict` structure

        Returns
        -------
        files : `dict` of `dicts`
            `dict` of ``(channel, fdict)`` pairs, where ``fdict`` is a `dict`
            of ``(fileformat, filenames)`` pairs.
        """
        segments = self.output_segments(start, end)
        fileformats = self.output_formats()
        extension = {
            'root': 'root',
            'txt': 'txt',
            'xml': 'xml',
            'hdf5': 'h5',
        }

        # build list of files
        outdir = self.get('OUTPUT', 'DIRECTORY')
        out = {}
        for channel in self.getlist('DATA', 'CHANNELS'):
            cstr = CHANNEL_DELIM_REGEX.sub('_', channel).replace('_', '-', 1)
            out[channel] = dict((form, []) for form in fileformats)
            for seg in segments:
                basename = '%s_OMICRON-%d-%d' % (cstr, seg[0], abs(seg))
                for form in fileformats:
                    out[channel][form].append(os.path.join(
                        outdir, channel,
                        '{0}.{1}'.format(basename, extension[form]),
                    ))
        if flatten:  # return a basic list of filenames
            return [f for c in out for form in out[c] for f in out[c][form]]
        return out



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2016)
#
# This file is part of PyOmicron.
#
# PyOmicron is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyOmicron is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyOmicron.  If not, see <http://www.gnu.org/licenses/>.

"""Constants and variables for Omicron processing
"""

import os
import re
from pathlib import Path

IFO_FQDN_REGEX = {
    "G1": re.compile(r"\.uni-hannover"),
    "H1": re.compile(r"\.ligo-wa"),
    "K1": re.compile(r"\.kagra"),
    "L1": re.compile(r"\.ligo-la"),
    "V1": re.compile(r"\.(virgo|ego-gw)"),
}


# -- generic parameters
try:
    IFO = os.environ['IFO']
except KeyError:
    from socket import getfqdn
    fqdn = getfqdn()
    for _key, _regex in IFO_FQDN_REGEX.items():
        if _regex.search(fqdn):
            IFO = _key.upper()
            break
    else:
        IFO = None
    ifo = os.getenv('ifo', IFO.lower() if IFO else None)
else:
    ifo = os.getenv('ifo', IFO.lower())
SITE = os.getenv('SITE')
site = os.getenv('site', SITE and SITE.lower() or None)

# -- omicron directories
HOME = Path.home()
# where Omicron runs
OMICRON_BASE = HOME / "omicron"
# where Omicron triggers are produced
OMICRON_PROD = OMICRON_BASE / "online"
# archive storage directory
OMICRON_ARCHIVE = HOME / "triggers"
# tag Omicron itself places on XML files
OMICRON_FILETAG = 'OMICRON'

# omicron channel files
if ifo is not None:
    OMICRON_GROUP_FILE = OMICRON_PROD / f"{ifo}-groups.txt"
    OMICRON_CHANNELS_FILE = OMICRON_PROD / f"{ifo}-channels.ini"
else:
    OMICRON_GROUP_FILE = None
    OMICRON_CHANNELS_FILE = None



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2016)
#
# This file is part of PyOmicron.
#
# PyOmicron is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyOmicron is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyOmicron.  If not, see <http://www.gnu.org/licenses/>.

"""Miscellaneous utilities
"""

import os
import subprocess
import sys
from pathlib import Path
from shutil import which
from tempfile import gettempdir

from packaging.version import Version

from . import const


def get_output_directory(args):
    """Return the output directory as parsed from the command-line args
    """
    return str(get_output_path(args))


def get_output_path(args):
    """Return the output path as parsed from the command-line args
    """
    if args.output_dir is None:
        if args.gps is None:
            dirname = args.group
        else:
            dirname = "{0}-{1[0]}-{1[1]}".format(args.group, args.gps)
        return (const.OMICRON_PROD / dirname).resolve(strict=False)
    return args.output_dir.resolve(strict=False)


def find_omicron():
    """Find the omicron executable in the environment

    Either via `PATH` or relative to the current python interpreter

    Returns
    -------
    path : `pathlib.Path`
        the path of the omicron executable

    Raises
    ------
    RuntimeError
        if omicron cannot be found, or is not executable
    """
    exe = which(
        "omicron",
        path=os.pathsep.join((
            os.getenv("PATH", ""),
            str(Path(sys.executable).parent),
        )),
    )
    if not exe or not os.access(exe, os.X_OK):
        raise RuntimeError(
            "cannot locate omicron in environment or is not executable"
        )
    return Path(exe).resolve()


def get_omicron_version(executable=None):
    """Parse the version number from the Omicron executable path

    Parameters
    ----------
    executable : `str`
        path of Omicron executable

    Returns
    -------
    version : `str`
        the Omicron-format version string, e.g. `vXrY`

    Examples
    --------
    >>> get_omicron_version()
    '2.1.0'
    """  # noqa: E501
    executable = executable or find_omicron()
    try:
        return Version(
            subprocess.check_output(
                [executable, "version"],
                env={"OMICRON_HTML": gettempdir()},
            ).decode("utf-8").rsplit(maxsplit=1)[-1],
        )
    except subprocess.CalledProcessError as ex:
        raise RuntimeError(f"failed to determine omicron version from executable: {str(ex)}")


def astropy_config_path(parent, update_environ=True):
    """Create and return a directory for a temporary astropy config path
    """
    parent = Path(parent)
    astropath = parent / ".config" / "astropy"
    astropath.mkdir(exist_ok=True, parents=True)
    confpath = astropath.parent
    if update_environ:
        os.environ["XDG_CONFIG_HOME"] = str(confpath)
    return confpath



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2019)
#
# This file is part of PyOmicron.
#
# PyOmicron is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyOmicron is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyOmicron.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for omicron.const
"""

import os
from importlib import reload
from unittest import mock

import pytest

from .. import const


@pytest.mark.parametrize("fqdn, ifo", [
    ("ldas-grid.ligo-wa.caltech.edu", "H1"),
    ("k1sum0.kagra", "K1"),
    ("test.localhost", None)
])
@mock.patch("socket.getfqdn")
def test_ifo(getfqdn, fqdn, ifo):
    getfqdn.return_value = fqdn
    reload(const)
    assert const.IFO == ifo
    assert const.ifo == (ifo.lower() if ifo else None)


def test_omicron_vars():
    assert str(const.OMICRON_BASE) == os.path.join(
        os.environ["HOME"],
        "omicron",
    )
